// ---------------------------------------------------------------------------
// Created by F. Malpartida on 9/9/12.
// Copyright 2011 - Under creative commons license 3.0:
//        Attribution-NonCommercial-ShareAlike CC BY-NC-SA
//        http://www.electrofunltd.com
// 
// This software is furnished "as is", without technical support, and with no 
// warranty, express or implied, as to its usefulness for any purpose.
//
// Thread Safe: No
// Extendable: No
//
// @file mmi.h
// Main application MMI header files. 
// 
// @brief 
// Requires New LiquidCrystal library to be installed in system:
//    https://bitbucket.org/fmalpartida/new-liquidcrystal/wiki/Home
// Renders to an LCD module the main status of the CNC controller: axis position
// feed rates, on board temperature, fan speeds.
// MMI availability can be compiled out from the system by undifining the _MMI_ENABLED_
// flag from the configMe.h header file.
//
//
// @author F. Malpartida
// ---------------------------------------------------------------------------
#ifndef _MMI_H_
#define _MMI_H_
#include "configMe.h"

#ifdef _MMI_ENABLED_
#include <LCD.h>
#include "tempCntrl.h"
#include "stepperCntrl.h"


class mmi 
{
public:
   
   /*!
    @function   
    @abstract   MMI class constructor.
    @discussion Creates and manages the MMI of the CNC machine. It requires to operate
    an LCD where to display the information, a temperature controller from this
    project and the stepper controller that will be used to display and extract positioning
    information.
    This class uses the New liquidCrystal library to drive the LCD that can be found in:
    https://bitbucket.org/fmalpartida/new-liquidcrystal/wiki/Home
    @param      none
    */  
   mmi ( void );
   
#ifdef _TEMP_CONTROL_ENABLED_
   /*!
    @function
    @abstract   Initializes the MMI. It requires an initialized LCD object reference,
    an initialized temperature controller and the stepper controller where it will extract
    positioning information, feed rates, etc to display on screen.
    @discussion 
    @param      lcd: New LiquidCrystal library LCD abstract object, initilized to use
    a particular LCD driver module.
    @param      temperatureController: a temperature controller used to monitor the onboard
    temperature and activate a fan. The MMI will extract from the controller temperature readings
    and fan velocity.
    @param       controller: initialized stepper controller driving the machine. This controller
    will be used to display the current axis positions, feedrates, etc.
    */  
   void init ( LCD *lcd, tempCntrl *temperatureController, stepperCntrl *controller );
#else   
   /*!
    @function
    @abstract   Initializes the MMI. It requires an initialized LCD object reference,
    an initialized temperature controller and the stepper controller where it will extract
    positioning information, feed rates, etc to display on screen.
    @discussion 
    @param      lcd: New LiquidCrystal library LCD abstract object, initilized to use
    a particular LCD driver module.
    @param       controller: initialized stepper controller driving the machine. This controller
    will be used to display the current axis positions, feedrates, etc.
    */  
   void init ( LCD *lcd, stepperCntrl *controller );
#endif // _TEMP_CONTROL_ENABLED_
   
   /*!
    @function   
    @abstract   main LCD rendering routine.
    @discussion Main rendering routine for the built in MMI. This method has to be called periodically
    in the main application thread or loop to refresh the LCD.
    */
   void display ( void );
   
private:
   LCD          *_lcd;
#ifdef _TEMP_CONTROL_ENABLED_
   tempCntrl    *_tempCont; 
#endif // _TEMP_CONTROL_ENABLED_
   stepperCntrl *_machCntrl;
   bool _init;
};

#endif // _MMI_ENABLED_
#endif
