// ---------------------------------------------------------------------------
// Created by F. Malpartida on 9/9/12.
// Copyright 2011 - Under creative commons license 3.0:
//        Attribution-NonCommercial-ShareAlike CC BY-NC-SA
//        http://www.electrofunltd.com
// 
// This software is furnished "as is", without technical support, and with no 
// warranty, express or implied, as to its usefulness for any purpose.
//
// Thread Safe: No
// Extendable: No
//
// @file stepperCntrl.cpp
// Main machine controller. 
// 
// @brief This class does all the coordinated movement of the stepper motors in
// terms of moving the three axis to a given coordinate. This class controls
// the linear movements, controls the machine's coordinates and positioning, 
// axis velocity.
//
// Acknowledgements:
// Inspired on the work of Carlos guilarte y diego colonnello (Txapuzas CNC).
//
// @author F. Malpartida
// ---------------------------------------------------------------------------
#include <stdlib.h>
#include <math.h>
#include <Arduino.h>
#include "configMe.h"
#include "stepperCntrl.h"
#include "stepper.h"


// CONSTANTS AND DEFINITIONS
// ---------------------------------------------------------------------------


// CLASS CONSTRUCTORS
// ---------------------------------------------------------------------------
// Constructor
stepperCntrl::stepperCntrl ( void )
{
   
   // Create all the stepper drivers that the machine will use
   _xStepper = new stepper ( X_STEP_PIN, X_DIR_PIN, X_ENABLE_PIN, 5 );
   _yStepper = new stepper ( Y_STEP_PIN, Y_DIR_PIN, Y_ENABLE_PIN, 5 );
   _zStepper = new stepper ( Z_STEP_PIN, Z_DIR_PIN, Z_ENABLE_PIN, 5 );
   
   // Disable all the drivers
   disableSteppers ();
   
   // Initialise positions, feedrates and defaults
   _units_in_mm    =  1.0;            // Start with mm - default
   _stepsPerUnit.x =  X_STEPS_PER_MM;
   _stepsPerUnit.y =  Y_STEPS_PER_MM;
   _stepsPerUnit.z =  Z_STEPS_PER_MM;
   
   _current_units.x = 0.0;
   _current_units.y = 0.0;
   _current_units.z = 0.0;
   
   _target_units.x = 0.0;
   _target_units.y = 0.0;
   _target_units.z = 0.0;
  
   _delta_units.x = 0.0;
   _delta_units.y = 0.0;
   _delta_units.y = 0.0;
   
   _delta_steps.x = 0;
   _delta_steps.y = 0;
   _delta_steps.y = 0;
   
   _direction.x = 0;
   _direction.y = 0;
   _direction.z = 0;
   
   _lastSpeed     = 0;                // Last speed set
   _feedrateDelay = 0;
   _feedRate      = FAST_XY_FEEDRATE;
   _curveSection  = CURVE_SECTION_MM;
   _absMode       = true;             // Work in absolute mode by default
   
   computeDeltas();
}


// PUBLIC METHODS
// ---------------------------------------------------------------------------
void stepperCntrl::init ( void )
{
   _feedrateDelay = calculateFeedrateDelay(FAST_XY_FEEDRATE);
}


void stepperCntrl::enableSteppers ( void )
{
   _xStepper->enable();
   _yStepper->enable();
   _zStepper->enable();   
}

void stepperCntrl::disableSteppers ( void )
{
   _xStepper->disable();
   _yStepper->disable();
   _zStepper->disable();   
}

void stepperCntrl::computeDeltas (void)
{
   //figure our deltas.
   _delta_units.x = abs(_target_units.x - _current_units.x);
   _delta_units.y = abs(_target_units.y - _current_units.y);
   _delta_units.z = abs(_target_units.z - _current_units.z);
   
   //set our steps current, target, and delta
   _current_steps.x = toSteps(_stepsPerUnit.x, _current_units.x);
   _current_steps.y = toSteps(_stepsPerUnit.y, _current_units.y);
   _current_steps.z = toSteps(_stepsPerUnit.z, _current_units.z);
   
   _target_steps.x = toSteps(_stepsPerUnit.x, _target_units.x);
   _target_steps.y = toSteps(_stepsPerUnit.y, _target_units.y);
   _target_steps.z = toSteps(_stepsPerUnit.z, _target_units.z);
   
   // separate the number of steps from its direction
   _delta_steps.x = abs(_target_steps.x - _current_steps.x);
   _delta_steps.y = abs(_target_steps.y - _current_steps.y);
   _delta_steps.z = abs(_target_steps.z - _current_steps.z);
   
   // what is our direction
   _direction.x = _target_units.x >= _current_units.x;
   _direction.y = _target_units.y >= _current_units.y;
   _direction.z = _target_units.z >= _current_units.z;
}


void stepperCntrl::setPosition(float x, float y, float z)
{
   _current_units.x = x;
   _current_units.y = y;
   _current_units.z = z;
   
   // compute the movement that would have to happen
   computeDeltas();
}

void stepperCntrl::setTarget(float x, float y, float z)
{
   // Set new position based on configuration
   // ---------------------------------------
   if ( _absMode )
   {
      _target_units.x = x;
      _target_units.y = y;
      _target_units.z = z;
   }
   else
   {
      _target_units.x = _current_units.x + x;
      _target_units.y = _current_units.y + y;
      _target_units.z = _current_units.z + z;
   }
   computeDeltas();
   setFeedDelay ();
}

stepperCntrl::t_floatPoint stepperCntrl::getPosition ( void )
{
   return (_current_units );
}

void stepperCntrl::setUnits (float units)
{
   _units_in_mm = units;
}

void stepperCntrl::moveLine(void)
{
   long x_counter, y_counter, z_counter, max_delta;
   uint16_t milliDelay = 0;
   
   //our step flags
   bool x_can_step = false;
   bool y_can_step = false;
   bool z_can_step = false;
   
   
   //figure out our deltas - the dominant delta
   max_delta = max(_delta_steps.x, _delta_steps.y);
   max_delta = max(_delta_steps.z, max_delta);
   
   x_counter = -max_delta/2;
   y_counter = -max_delta/2;
   z_counter = -max_delta/2;
   
   if (_feedrateDelay >= 16383)
      milliDelay = _feedrateDelay / 1000;
   
   //enable our steppers
   enableSteppers ();
   
   // set stepper directions
   _xStepper->setDirection (stepper::t_dir(_direction.x));
   _yStepper->setDirection (stepper::t_dir(_direction.y));
   _zStepper->setDirection (stepper::t_dir(_direction.z));
   
   //do our DDA line!
   do
   {
      // Check that we can move in all directions
      x_can_step = canStep(X_MIN_PIN, X_MAX_PIN, _current_steps.x, _target_steps.x, 
                           _direction.x);
      y_can_step = canStep(Y_MIN_PIN, Y_MAX_PIN, _current_steps.y, _target_steps.y, 
                           _direction.y);
      z_can_step = canStep(Z_MIN_PIN, Z_MAX_PIN, _current_steps.z, _target_steps.z, 
                           _direction.z);
      
      // If we can move, move each axis according to computed values for the line
      // ------------------------------------------------------------------------
      if (x_can_step)
      {
         x_counter += _delta_steps.x;
         
         if (x_counter > 0)
         {
            _xStepper->step();
            x_counter -= max_delta;
            
            if (_direction.x)
               _current_steps.x++;
            else
               _current_steps.x--;
         }
      }
      
      if (y_can_step)
      {
         y_counter += _delta_steps.y;
         
         if (y_counter > 0)
         {
            _yStepper->step();
            y_counter -= max_delta;
            
            if (_direction.y)
               _current_steps.y++;
            else
               _current_steps.y--;
         }
      }
      
      if (z_can_step)
      {
         z_counter += _delta_steps.z;
         
         if (z_counter > 0)
         {
            _zStepper->step();
            z_counter -= max_delta;
            
            if (_direction.z)
               _current_steps.z++;
            else
               _current_steps.z--;
         }
      }
      
      // wait for next step according to the computed feed delay
      // -------------------------------------------------------
      if (milliDelay > 0)
      {
         delay(milliDelay);			         
      }
      else
      {
         delayMicroseconds(_feedrateDelay);         
      }
   }
   while (x_can_step || y_can_step || z_can_step);
   
   //set our current position and target position to be the same
   _current_units.x = _target_units.x;
   _current_units.y = _target_units.y;
   _current_units.z = _target_units.z;
   computeDeltas();
}

void stepperCntrl::moveArc ( t_floatPoint orig, double i, double j, bool cwMove )
{
   stepperCntrl::t_floatPoint cent, a, b, newPoint;
   float angleA, angleB, angle, radius, length;
   int steps, s, step;
   
   cent = getPosition();
   
   // Center coordinates are always relative
   cent.x += i;
   cent.y += j;
   
   a = getPosition();
   
   a.x -= cent.x;
   a.y -= cent.y;
   b.x = (orig.x - cent.x);
   b.y = (orig.y - cent.y);
   
   // Clockwise move ?
   if (cwMove) 
   { 
      angleA = atan2(b.y, b.x);
      angleB = atan2(a.y, a.x);
   } 
   else 
   { // Counterclockwise
      angleA = atan2(a.y, a.x);
      angleB = atan2(b.y, b.x);
   }
   
   // Make sure angleB is always greater than angleA and if not add 2PI so that 
   // it is (this also takes care of the special case of angleA == angleB,
   // ie we want a complete circle)
   // -------------------------------------------------------------------------
   if (angleB <= angleA)
   {
      angleB += 2 * M_PI;
   }
   
   angle = angleB - angleA;
   
   radius = sqrt(a.x * a.x + a.y * a.y);
   length = radius * angle;
   
   steps = (int) ceil(length / _curveSection);
   
   for (s = 1; s <= steps; s++) 
   {
      step = (!cwMove) ? s : steps - s; // Work backwards for CW
      newPoint.x = cent.x + radius * cos(angleA + angle * ((float) step / steps));
      newPoint.y = cent.y + radius * sin(angleA + angle * ((float) step / steps));
      
      setTarget(newPoint.x, newPoint.y, orig.z);
      
      // Move
      moveLine();
   }
}

long stepperCntrl::getMaxSpeed(void)
{
   // Calculate the max feedrate. If there is a value for the Z axis, use
   // this value first
   // --------------------------------------------------------------------
   if (_delta_steps.z > 0)
   {
      return calculateFeedrateDelay(FAST_Z_FEEDRATE);
   }
   else
   {
      return calculateFeedrateDelay(FAST_XY_FEEDRATE);
   }
}

void stepperCntrl::setFeedRate ( long feedrate )
{
   _feedRate = feedrate;
}

long stepperCntrl::getFeedRate ( void )
{
   return ( _feedRate );
}

// Setters and getters.
void stepperCntrl::setCurveSection ( float curveSection )
{
   _curveSection = curveSection;
}

float stepperCntrl::getCurveSection ( void )
{
   return (_curveSection);
}

void  stepperCntrl::setAbsoluteCoord ( void )
{
   _absMode = true;
}

void  stepperCntrl::setRelativeCoord ( void )
{
   _absMode = false;
}

bool stepperCntrl::absoluteCoord ( void )
{
   return ( _absMode );
}

// PRIVATE METHODS
// ---------------------------------------------------------------------------

long stepperCntrl::toSteps ( float stepsPerUnit, float units )
{
   return (( stepsPerUnit * units * _units_in_mm )/SCALE_STEPS);
}

bool stepperCntrl::canStep(uint8_t min_pin, uint8_t max_pin, long current, 
                           long target, uint8_t direction)
{
   // stop us if we're on target
   // -------------------------
   if (target == current)
   {
      return false;
   }
   // stop us if we're at home (hit the switch) and still going 
   else if (readSwitch(min_pin) && !direction)
   {
      return false;
   }
   // stop us if we're at max (hit the switch) and still going
   else if (readSwitch(max_pin) && direction)
   {
      return false;
   }   
   //default to being able to step
   return true;
}


bool stepperCntrl::readSwitch(uint8_t pin)
{
   // If the end stop switch is configured, read it to see if we are at the end.
   if ( pin != 0 )
   {
      if ( SENSORS_INVERTING )
         return !digitalRead(pin);
      else
         return digitalRead(pin);
   }
   return (false);
}

long stepperCntrl::calculateFeedrateDelay (long feedrate)
{
   long retVal = 0;
   long maxSteps;
   
   //how long is our line length?
   float distance = sqrt(_delta_units.x * _delta_units.x + 
                         _delta_units.y * _delta_units.y + 
                         _delta_units.z * _delta_units.z);
   
   //find the dominant axis.
   maxSteps = max(_delta_steps.x, _delta_steps.y);
   maxSteps = max(_delta_steps.z, maxSteps);
   
   // calculate delay between steps in microseconds.
   // the formula has been condensed to save space.  Here it is in english:
   // distance / feedrate * 60000000.0 = move duration in microseconds
   // move duration / maxSteps = time between steps for master axis.
   if ( maxSteps != 0 )
   {
      retVal = (((distance * _units_in_mm) * 600000000.0) / feedrate) / maxSteps;
   }
   return (retVal);
}

void stepperCntrl::setFeedDelay ( void )
{
   if (_feedRate > 0)
   {	
      _feedrateDelay = calculateFeedrateDelay(_feedRate);
      if ( _feedrateDelay > 0 )
      {
         _lastSpeed = _feedrateDelay;
      }
   }
   // nope, no feedrate configure yet
   else
   {
      // If we had a lastSpeed at which we moved use it. If not use the max
      // feed rate.
      if (_lastSpeed > 0)
      {  
         _feedrateDelay = _lastSpeed;
      }
      else
      {
         _feedrateDelay = getMaxSpeed();
      }  
   }
}


